QuestHubber = LibStub("AceAddon-3.0"):NewAddon("QuestHubber", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0");
local core = QuestHubber;
-- local L = LibStub("AceLocale-3.0"):GetLocale("QuestHubber");
local G = LibStub:GetLibrary("LibGratuity-3.0");

--[[local options = {
	name = "QuestGather",
	type = 'group',
	args = {
	},
};]]
local defaults = {
	profile = {
		debug = 0,
	},
};
local FACTION_MAP = {
	Alliance = 1,
	Horde = 2,
};

-- TODO (possibly)
-- package scripts (on ignore)
-- setup pkgmeta for multiple addons
-- change the way unknown quests show
-- add checkbox in unknown quests area
-- object tooltip scan
-- make tooltips with multiple quests look consistent
-- item drop quests

core.dataLoaded = false;		-- flag set after LoadStep() is done
core.currentZone = -1;			-- current World Map zone id
core.pinReg = {};				-- registry of all used pins
core.frameReg = {};				-- registry of all unused pin frmaes
core.playerQuests = {};			-- quests the player has completed, queried from server
core.showDaily = false;			-- flag to show daily quests (TODO: implement as option)
core.questLink = {				-- table linking optional quests to the mandatory quests they lead to which take away the initial
	[25985] = 27874,
	[27727] = 27203,
};
core.questLog = {};				-- current quest log, used to see when quests are accepted or completed
core.questInit = false;			-- flag set after the quest log has been initially recorded
core.abandonQuest = -1;			-- when "Abandon" is clicked, sets the quest that was selected at the time; used after confirmation
core.cache = {};				-- cache for some quick lookups

function core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("QuestHubberDB", defaults, "Default");
	self:RegisterChatCommand("qhub", function()
		self:Print("Initializing loader...")
		self:InitializeLoader()
	end)
	--[[
		LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("", options);
		local ACD = LibStub("AceConfigDialog-3.0");
		ACD:AddToBlizOptions("", "");
		self:SetEnabledState(self.db.profile.enabled;
	]]
	
	self.cache = {
		quests = {},		-- [questid] = "Localized Name",
		npcs = {},			-- [npcid] = "Localized Name",
		-- zones = {},			-- [zoneid] = "Localized Name",  -- NOT USED
		invalid = {},		-- [npcid] = true, -- flag set so you don't get double debug warnings about a single npc
		schedule = {},		-- [questid] = true, -- scheduled for rescan due to 
	};
	
	--[[ start the loader if we have data; otherwise, wait for QuestGather
	if (not self.data or not self.data.quests) then
		self.data = {};
		self.data.quests = {};
	else
		--self:InitializeLoader();
	end]]
	
	self.data = {
		zones = {},
		quests = {}, -- unused
	};
end

-- unused?
local stepSpace = 0.05 				-- time to spend in seconds
function core:InitializeLoader()
	-- this is called when we have data
	self.loader = CreateFrame("Frame");
	local i = 0;
	self.loader:SetScript("OnUpdate", function(_, elapsed)
		i = i + elapsed;
		if (i > stepSpace) then
			self:LoadStep();
		end
	end);
end

function core:OnEnable()	
	self:RegisterEvent("WORLD_MAP_UPDATE"); 			-- map change
	self:RegisterEvent("QUEST_QUERY_COMPLETE"); 		-- get quests completed
	
	self:RegisterEvent("QUEST_LOG_UPDATE");				-- scan for quest completion/pickup
	
	self:Hook("SetAbandonQuest", true);					-- hook for abandoned quests
	self:Hook("AbandonQuest", true);					-- hook for confirmation
	
	self:QueryQuestsCompleted();
end

function core:OnDisable()
	
end

function core:SlashHandle(message)

end

-- currently unused
local npc = {"begin", "complete"};	-- table for looping through npcs in each quest
local lastStep = nil;				-- iterator position
local stepTime = 0.02; 				-- time to spend in seconds
function core:LoadStep()
	local start = GetTime();
	local steps = 0;
	-- custom iterator that picks up from where the last LoadStep left off
	for q, v in next, self.data.quests, lastStep do
		if (tonumber(q)) then	-- make sure it's a fully cached quest
			-- Load Step
			steps = steps + 1
			
			for i, key in ipairs(npc) do
			
				-- cache npc id -> name
				if (v[key] and v[key]["npcid"]) then
					-- read from tooltip
					local id = v[key]["npcid"];
					-- check cache
					local name = self.cache.npcs[id];
					if (not name) then
						self:NPCScan(id);
					end
					if (name) then
						-- update name
						v[key]["npc"] = name;
					end
				end
				
				--[[ cache zone id -> name
				if (v[key] and v[key]["zoneid"]) then
					local id = v[key]["zoneid"];
					local name = self.cache.zones[id];
					
					if (not name and tonumber(id)) then	
						SetMapByID(id);
						name = GetMapInfo();
					end
					if (name) then
						self.cache.zones[id] = name;
						-- also setting english name -> id cache
						-- if (v[key]["zone"]) then
							-- self.cache.engZones[(v[key]["zone"])] = id;
						-- end 
					end
				end]]
				
				-- set data by zone id
				if (v[key] and v[key]["zoneid"]) then
					local id = v[key]["zoneid"];
					if (id == 301) then
						-- self:Print(q);
					end
					if (not self.data.zones[id]) then
						self.data.zones[id] = {};
					end
					
					self.data.zones[id][q] = v;
				end
			end
			
			if (GetTime()-start >= stepTime) then
				self:Debug("Breaking:", steps, "steps; ", ("%.3fs"):format(GetTime()-start))
				break;
			end
		end
		
		lastStep = q;
	end
	
	
	if (not next(self.data.quests, lastStep)) then
		--[[ temp test
		self.data.zones[301]={} -- Stormwind City
		for i=1,10 do
			self.data.zones[301][i^2] = {
				begin = {
					x = i/10,
					y = i/10,
				},
				name = "Test "..i,
			}
		end]]
		
		self.dataLoaded = true;
		self:Debug("Data loaded");
		self.loader:SetScript("OnUpdate", nil);
		self:QueryQuestsCompleted();
		self:UpdatePins(true);
	end
end

function core:QueryQuestsCompleted()
	QueryQuestsCompleted();
	self:ScheduleTimer(function()
		if (not next(self.playerQuests)) then
			self:Debug("Could not retrieve quest data, trying again.");
			self:QueryQuestsCompleted();
		end
	end, 10);
end

-- Unused?
function core:QueryQuestNames()
	self:Debug("QueryQuestNames START");
	local start = GetTime();
	local loop = 0;
	for i,v in pairs(self.cache.schedule) do
		if (i > 0) then
			loop = loop + 1;
			-- self:QuestScan(id);
			if (self.cache.quests[i]) then 
				self.cache.schedule[i] = nil
			end
		end
		if GetTime()-start > -1 then
			self:Debug(format("Breaking secondary lookup after %d loops; %.3fs", loop, start));
			break;
		end
	end
	
	if (next(self.cache.schedule)) then
		self:ScheduleTimer("QueryQuestNames", 20);
		self:Debug("Some quest names were not cached. Checking in 20s");
	end
	self:Debug("QueryQuestNames END");
end

function core:QuestScan(id)
	-- self:Debug("Scanning", id)
	if id and id > 0 then
		G:SetHyperlink(format("quest:%d", id));
		local text = G:GetText();
		if (text and text[1] and text[1][1]) then
			self.cache.quests[id] = text[1][1];
			-- self:Debug(format("Quest id %d = %q", q, text[1][1]))
		else
			-- self:Debug("No name for quest "..id);
			-- self.cache.schedule[id] = true
			self:ScheduleTimer(function()
				self:QuestScan(id)
			end, 2)
		end
	end
end

function core:NPCScan(id)
	G:SetHyperlink(format("unit:0xF53%05X00000000", id));
	local text = G:GetText();
	if not text and not self.cache.invalid[id] then
		-- self:Debug("Invalid NPC id:", id, "- perhaps not cached?");
		self.cache.invalid[id] = true;
	end
	if (text and text[1] and text[1][1]) then
		self.cache.npcs[id] = text[1][1];
		name = text[1][1];
	end
end

function core:WORLD_MAP_UPDATE()
	if (self.dataLoaded) then
		self:UpdatePins();
	end
end

-----------------------------------------------
-- Pin management
-----------------------------------------------

function core:UpdatePins(force)
	local id = GetCurrentMapAreaID();
	local noBeginCt = 0;
	if (id ~= self.currentZone or force) then
		-- update all pins
		self.currentZone = id;
		self:HideAllPins();
		if (self.data.zones[id]) then
			local total = 0;
			for q,v in pairs(self.data.zones[id]) do
				if (not self:QuestActive(q)) then	-- if we're not on the quest or have completed it
					self:ShowPin(q, v);
					total = total + 1;
					-- self:Debug("Showing pin",q,"on map",id)
				else
					-- self:Debug("Not showing pin",q,"on map",id)
				end
			end
			self:Debug("Showing",total,"pins");
		end
	end	
end

function core:ShowPin(id, data)
	local flags, level, xcoord, ycoord, npc, name = self:GetQuestData(data);
	if (not flags or not tonumber(xcoord) or not tonumber(ycoord)) then
		-- self:Debug("Bad quest data:", id, data, ";;;", flags, level, xcoord, ycoord, npc, name)
		return
	end
	
	-- check the cache for quest name; scan tooltip if not
	if (not self.cache.quests[id]) then
		self:QuestScan(id)
	end
	-- TODO: npc scan here as well
	
	local pin = self:GetPin();
	
	pin:ClearAllPoints();
	pin:SetParent(WorldMapDetailFrame);
	pin:SetPoint("CENTER", WorldMapDetailFrame, "TOPLEFT", xcoord/10000*WorldMapDetailFrame:GetWidth(), -ycoord/10000*WorldMapDetailFrame:GetHeight());
	pin:SetFrameStrata("FULLSCREEN");
	pin:SetFrameLevel(1000);
	pin:Show();
	
	self.pinReg[id] = pin;
end

function core:HidePin(id)
	if (self.pinReg[id]) then
		self.pinReg[id]:Hide();
		table.insert(self.frameReg, self.pinReg[id]);
		self.pinReg[id] = nil;
	end
end

function core:HideAllPins()
	for i,v in pairs(self.pinReg) do
		self:HidePin(i);
	end
end

function core:GetPin()
	if (#self.frameReg > 0) then
		return table.remove(self.frameReg);
	end
	
	-- create new
	local pin = CreateFrame("Frame");
	pin.texture = pin:CreateTexture();
	-- 256x64, 8x4
	pin.texture:SetTexture("Interface\\MINIMAP\\ObjectIcons.blp");
	pin.texture:SetTexCoord(0.125, 0.250, 0.250, 0.500);
	pin.texture:SetAllPoints();
	
	pin:SetWidth(16);
	pin:SetHeight(16);
	pin:EnableMouse(true);
	pin:SetScript("OnEnter", function(pin)
		self:ShowTooltip(pin);
	end);
	pin:SetScript("OnLeave", function()
		self:HideTooltip();
	end);
	
	return pin;
end

function core:QuestActive(id)
	if (not self.questLog[id] and not self.playerQuests[id] and self.questLink[id]) then
		return self.playerQuests[self.questLink[id]];
	end
	
	return self.questLog[id] or self.playerQuests[id];
end

-----------------------------------------------
-- Tooltip management
-----------------------------------------------

function core:GetMousePosition()
	local left, top = WorldMapDetailFrame:GetLeft(), WorldMapDetailFrame:GetTop();
	local width, height = WorldMapDetailFrame:GetWidth(), WorldMapDetailFrame:GetHeight();
	local scale = WorldMapDetailFrame:GetEffectiveScale();

	local x, y = GetCursorPosition();
	local cx = (x/scale - left) / width;
	local cy = (top - y/scale) / height;
	
	return math.min(math.max(cx, 0), 1), math.min(math.max(cy, 0), 1);
end

function core:Distance(x1, y1, x2, y2)
	-- gets distance in map units TODO: use yards?
	local distRatio = WorldMapDetailFrame:GetHeight() / WorldMapDetailFrame:GetWidth();
	
	return math.sqrt( (x1 - x2)^2 + ((y1 - y2)/distRatio)^2 );
end

function core:ShowTooltip(pin)
	GameTooltip:SetOwner(pin, "ANCHOR_TOPLEFT");
	GameTooltip:ClearLines();
	
	local first = true;	
	-- find all quests in range of hover
	local mx, my = self:GetMousePosition();
	for q,v in pairs(self.data.zones[self.currentZone]) do
		if (not self:QuestActive(q)) then
			local _, _, xcoord, ycoord, npc, npcname = self:GetQuestData(v);
			if (tonumber(xcoord) and tonumber(ycoord)) then
				local dist = self:Distance(mx, my, xcoord/10000, ycoord/10000);
				-- self:Print("d:",dist,v.begin.x,v.begin.y);
				if (dist <= 0.02) then
					if (not first) then
						GameTooltip:AddLine(" ");
					else
						first = false;
					end
					GameTooltip:AddLine(self.cache.quests[q] or "(Querying server...)");
					GameTooltip:AddLine(self.cache.npcs[npc] or npcname, 1, 1, 1, 1);
				end
			end
		end
	end
	
	GameTooltip:Show();
end

function core:HideTooltip()
	GameTooltip:Hide();
end

-----------------------------------------------
-- Quest log management
-----------------------------------------------

function core:QUEST_QUERY_COMPLETE()
	self.playerQuests = GetQuestsCompleted();
	self:UpdatePins(true);
end

function core:QUEST_LOG_UPDATE()
	-- get closed tabs
	local closed = {}
	for i=1,GetNumQuestLogEntries() do
		local title, _, _, _, isHeader, isCollapsed = GetQuestLogTitle(i);
		if (isCollapsed) then
			closed[title] = true
		end
	end
	ExpandQuestHeader(0)
	
	if (not self.questInit) then
		for i=1,GetNumQuestLogEntries() do
			local _, _, _, _, _, _, _, _, id = GetQuestLogTitle(i);
			if (id and id > 0) then
				self.questLog[id] = true;
			end
		end
		self:Debug("Quest Log snapshot saved");
		-- self:UnregisterEvent("QUEST_LOG_UPDATE");
		self.questInit = true;
	else
		-- self:Debug("QUEST_LOG_UPDATE");
		local tempLog = {};
		for i=1,GetNumQuestLogEntries() do
			local name, _, _, _, _, _, _, _, id = GetQuestLogTitle(i);
			-- self:Debug("   ", name, id);
			if (id and id > 0) then
				tempLog[id] = name;
			end
		end
		-- self:Debug("_END_");
		
		for i,v in pairs(self.questLog) do
			if (not tempLog[i]) then
				-- quest removed; hopefully completed
				self:Debug("Quest completed:",i);
				self.questLog[i] = nil;
				tempLog[i] = nil;
				self.playerQuests[i] = true;
				-- force a full update in case any optional pre-reqs should be turned off
				self:UpdatePins(true);
			end
		end
		
		-- merge added
		for i,v in pairs(tempLog) do
			if (not self.questLog[i]) then
				self:Debug("Quest accepted:", v);
				self.questLog[i] = true;
				self:HidePin(i);
			end
		end
	end
	
	-- restore closed tabs
	for i=1,GetNumQuestLogEntries() do
		local title, _, _, _, isHeader = GetQuestLogTitle(i);
		if closed[title] then
			CollapseQuestHeader(i)
		end
	end
end

function core:SetAbandonQuest()
	self.abandonQuest = GetQuestLogSelection();
end

function core:AbandonQuest()
	-- saved quest was abandoned
	-- self:Debug("AbandonQuest");
	local name, _, _, _, _, _, _, _, id = GetQuestLogTitle(self.abandonQuest);
	self:Debug("Quest abandoned:",name);
	self.questLog[id] = nil;
	local quest = self.data.quests[id];
	if (quest and quest.begin and quest.begin.zoneid == self.currentZone) then
		self:ShowPin(id, quest, "begin");
	end
end

function core:Debug(level, ...)
	if (type(level) ~= "number") then
		self:Debug(1, level, ...);
		return;
	end
	if self.db.profile.debug >= level then 
		self:Print(...);
	end
end

local loader  -- timer that runs UpdatePins after everything is loaded
function core:RegisterQuests(zones)
	for i,v in pairs(zones) do
		if (not self.data.zones[i]) then
			self.data.zones[i] = {};
		end
		
		for q,qv in pairs(v) do
			self.data.zones[i][q] = qv;
		end
	end
	self.dataLoaded = true;
	if (not loader) then
		loader = self:ScheduleTimer(function()
			self:UpdatePins(true);
		end, 1);
	end
end

-- returns flags, level, xcoord, ycoord, npc, npcname
function core:GetQuestData(questStr)
	return questStr:match("(%d+):(%d+):(%d+):(%d+):(%d+):(.+)")
end